/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package GUI;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import util.*;

/*********************************************************
* Class for catching events when user pushes menu items
* Actions include open, save, save as, and exit.
*
* @author Olav Haugan
*
* Date: 14.10.01
*********************************************************/
public class Action implements ActionListener
{
    private MQCFrame parent;
    private File currentFile;
    private File currentDir;
    
    private static final String EXTENSION = "mqcc";
    private static final String EXT_DESC = "MQC configuration (*.mqcc)";    
    private static final String ABOUT_INFO1 = "MicroQoSCORBA configuration tool v. 1.0";
    private static final String ABOUT_INFO2 = "Olav Haugan 2001";
    private static final String ABOUT_INFO3 = "Washington State University";    
                
    /*********************************************************
    * Constructor. Initializes variables
    *
    * @param MQCFrame The frame that contains menu and buttons
    *********************************************************/
    public Action( MQCFrame parent )
    {
        this.parent = parent;
        currentFile = null;
        currentDir = null;
    }
    
    /*********************************************************
    * Returns the current selected file to save or load from.
    *********************************************************/
    public File getCurrentFile()
    {
        return currentFile;
    }
    
    /*********************************************************
    * Function is called by the underlying framework 
    * when a menu or button is pushed. The function calls the
    * appropriate handler function.
    *
    * @param ActionEvent The event that happens when a button
    *                    is pushed.
    *********************************************************/
    public void actionPerformed( ActionEvent ev )
    {
        String command = ev.getActionCommand();
        if(command.equals("Open...") || command.equals("Open") )
        {
            //Show open file dialog
            doOpen();
        }
        else if(command.equals("Save"))
        {
            //Show save file dialog         
            doSave();
        }
        else if(command.equals("Save as..."))
        {
            //Show save file dialog
            doSaveAs();
        }
        else if(command.equals("About"))
        {
            //Show about dialog
            doAbout();
        }
        else if(command.equals("Check profile"))
        {
            //Run profile checker
            doCheckProfile();
        }
        else if(command.equals("Exit"))
        {
            //Do exit
            System.exit(0);
        }
    }
    
    /*********************************************************
    * Shows a file chooser dialog. Records the selected 
    * directory and file.
    *********************************************************/
    public void doOpen()
    {
        JFileChooser fileChooser = new JFileChooser(System.getProperty("user.dir"));
        if(currentDir != null)
            fileChooser.setCurrentDirectory(currentDir);
        
        //Filter for selecting properties files
        ExtensionFileFilter filter = new ExtensionFileFilter();
        filter.addExtension(EXTENSION);
        filter.setDescription(EXT_DESC);
        fileChooser.addChoosableFileFilter( filter );
        
        int selected = fileChooser.showOpenDialog(parent);
        //Get selected directory
        currentDir = fileChooser.getCurrentDirectory();
        switch(selected)
        {
            case JFileChooser.APPROVE_OPTION:
                //Get selected file
                currentFile = fileChooser.getSelectedFile();
                //Load the properties from the file
                loadProperties();
                //Populate the controls in the GUI
                parent.populate();
                break;
            case JFileChooser.CANCEL_OPTION:
                currentFile = null;
                break;
            default:
                currentFile = null;
        }
    }

    /*********************************************************
    * Saves the properties to a file.
    * If no file is selected a file chooser dialog is shown.
    *********************************************************/
    public void doSave()
    {
        //No file selected
        if(currentFile == null)
        {
            doSaveAs();
        }
        else
        {
            //Save the properties to the selected file
            saveProperties();
        }
    }

    /*********************************************************
    * Shows a file chooser dialog. Records the selected 
    * directory and file. Properties are saved to the selected
    * file.
    *********************************************************/
    public void doSaveAs()
    {
        JFileChooser fileChooser = new JFileChooser(System.getProperty("user.dir"));
        if(currentDir != null)
            fileChooser.setCurrentDirectory(currentDir);
        
        //Filter for selecting properties files
        ExtensionFileFilter filter = new ExtensionFileFilter();
        filter.addExtension(EXTENSION);
        filter.setDescription(EXT_DESC);
        fileChooser.addChoosableFileFilter( filter );
        
        int selected = fileChooser.showSaveDialog(parent);
        //Get the current directory
        currentDir = fileChooser.getCurrentDirectory();
        switch(selected)
        {
            case JFileChooser.APPROVE_OPTION:
                //Get the current selected file
                currentFile = fileChooser.getSelectedFile();

                //Add extension to the file if none exists
                if(getExtension(currentFile) == null)
                {
                    StringBuffer fileName = new StringBuffer(currentFile.getPath());
                    fileName.append("." + EXTENSION);
                    File f = new File(fileName.toString());
                    currentFile = f;
                }               
                //Save the properties to the selected file
                saveProperties();
                break;
            case JFileChooser.CANCEL_OPTION:
                currentFile = null;
                break;
            default:
                currentFile = null;
        }
    }

    /*********************************************************
    * Shows an about dialog.
    *********************************************************/
    public void doAbout()
    {
        final JDialog dialog = new JDialog(new MQCFrame(), "About " + Constants.TOOL_TITLE);
        JPanel buttonPanel = new JPanel();
        JButton okBtn = new JButton("OK");
        
        //Create an action for the OK button
        okBtn.addActionListener( new ActionListener()
        {
            public void actionPerformed( ActionEvent ev )
            {
                dialog.dispose();   
            }
        });
        buttonPanel.add(okBtn);
        dialog.getContentPane().add( BorderLayout.SOUTH, buttonPanel );

        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        c.weightx = 0;
        c.gridwidth = GridBagConstraints.REMAINDER;
        c.anchor = GridBagConstraints.CENTER;
        
        JPanel textPanel = new JPanel(gridbag);
        JLabel label1 = new JLabel(ABOUT_INFO1);
        gridbag.setConstraints(label1, c);
        JLabel label2 = new JLabel(ABOUT_INFO2);
        gridbag.setConstraints(label2, c);
        JLabel label3 = new JLabel(ABOUT_INFO3);
        gridbag.setConstraints(label3, c);
                
        textPanel.add(label1);
        textPanel.add(label2);
        textPanel.add(label3);
        dialog.getContentPane().add(textPanel, BorderLayout.CENTER);
        dialog.setSize(330, 150);
        dialog.show();
    }

    /*********************************************************
    * Save the properties to the selected file.
    *********************************************************/
    private void saveProperties()
    {
        Properties props = parent.getProperties(true);
        try
        {
            //ArrayList myList = new ArrayList(props.values());
            //Collections.sort(myList);
            //System.out.println(props.toString());
            props.store(new FileOutputStream(currentFile), "MQC Config File");
        }
        catch(IOException ex)
        {
            ex.printStackTrace();
        }
    }

    /*********************************************************
    * Load the properties from the selected file.
    *********************************************************/
    private void loadProperties()
    {
        Properties props = parent.getProperties(false);
        try
        {
            props.load(new FileInputStream(currentFile));
        }
        catch(IOException ex)
        {
        }
    }
    
    /*********************************************************
    * Extracts a file's extension from its file name
    *
    * @param File The file that we want the extension from
    *********************************************************/
    private String getExtension( File f )
    {
        if(f != null)
        {
            String filename = f.getName();
            int i = filename.lastIndexOf( '.' );
            if( i > 0 && i < filename.length() - 1 )
            {
                return filename.substring(i+1).toLowerCase();
            }
        }
        return null;
    }
    
    /*********************************************************
    * Checks the current profile for ambiguity
    *********************************************************/
    private void doCheckProfile()
    {
        ProfileChecker profiler = new ProfileChecker(parent.getProperties(true));
        if(profiler.isProfileValid())
        {
            parent.setCheckProfileIcon(true);
        }
        else
        {
            showErrorDialog(profiler.getErrors());
            parent.setCheckProfileIcon(false);
        }
    }

    /*********************************************************
    * Show a dialog with errors
    *
    * @param Vector A vector of strings
    *********************************************************/
    private void showErrorDialog(Vector errors)
    {
        final JDialog dialog = new JDialog(new MQCFrame(), "Configuration error");
        JPanel buttonPanel = new JPanel();
        JButton okBtn = new JButton("OK");
        
        //Create an action for the OK button
        okBtn.addActionListener( new ActionListener()
        {
            public void actionPerformed( ActionEvent ev )
            {
                dialog.dispose();   
            }
        });
        buttonPanel.add(okBtn);
        dialog.getContentPane().add( BorderLayout.SOUTH, buttonPanel );
        JTextArea textArea = new JTextArea(errors.size(), 35);
        textArea.setEditable(false);
        textArea.setLineWrap(true);
        textArea.setWrapStyleWord(true);
        
        Enumeration e = errors.elements();
        String error;
        while(e.hasMoreElements())
        {
            error = (String) e.nextElement();
            textArea.append(error);
            textArea.append("\r\n");
        }
        dialog.getContentPane().add(textArea, BorderLayout.CENTER);
        dialog.pack();
        dialog.show();
    }
}
