/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

//include(`Config_macros.m4')

package mqc;

import mqc.ior.*;
import mqc.holders.*;
import mqc.transports.*;
import mqc.protocols.*;

import java.util.*;

/**
* Portable Object Adapter 
* 
* @author Olav Haugan
* @author Kim Christian Swenson
*
* 14.10.01
*/
public class POA
{
   private String name;
   private S_ORB orb;
   private int key;
   private Hashtable activeObjectMap;
   private Hashtable replyHandlers;

   /**
    * Constructor. Initializes variables. Adds itself to the
    * map of POA's in the ORB.
    *
    * @param orb  A reference to the orb
    * @param  The name of the POA
    */
   public POA( S_ORB _orb, String _name )
   {
      orb = _orb;
      name = _name;
      key = 0;
      activeObjectMap = new Hashtable(1);
      orb.addPOA(name, this);
      init();
   }

   /**
    * Initializes protocol reply message hashtable
    */
   public void init()
   {
      int noProtocols = Config.protocols.length;
      int prot = 0;

      replyHandlers = new Hashtable(noProtocols);
      for(int i = 0; i < noProtocols; i++)
      {
         prot = Config.protocols[i];
         switch(prot)
         {
         __GIOP_PROTOCOL__(
         case Config.GIOP:
            __GIOP10__( 
            replyHandlers.put("GIOP", new GIOPReply( 
                  new GIOPHeader_1_0(), 
                  new GIOPReplyHeader_1_0() )); 
            )
            __GIOP11__( 
            replyHandlers.put("GIOP", new GIOPReply( 
                  new GIOPHeader_1_1(), 
                  new GIOPReplyHeader_1_1() )); 
            )
            __GIOP12__( 
            replyHandlers.put("GIOP", new GIOPReply( 
                  new GIOPHeader_1_2(), 
                  new GIOPReplyHeader_1_2() ));
            )
            __GIOP13__( 
            replyHandlers.put("GIOP", new GIOPReply( 
                  new GIOPHeader_1_3(), 
                  new GIOPReplyHeader_1_3() ));
            )
            break;
         )
         __GIOPLITE_PROTOCOL__(
         case Config.GIOPLite:
            replyHandlers.put("GIOPLite", new GIOPLiteReply());
            break;
         )
         __MQCIOP_PROTOCOL__(
         case Config.MQCIOP:
            replyHandlers.put("MQCIOP", new MQCIOPReply());
            break;
         )
         default:
            throw new RuntimeException("No protocols specified");
         }
      }
   }

   /**
    * Receives a request from the ORB and dispatches the
    * request to the correct implementation servant.
    *
    *  @param msg  request to be handled
    *  @return  reply message (GIOP|MQC|GIOPLite)
    */
   public Reply handleRequest( Request msg )
   {
      //Find the correct Servant
      //See the idl generated xxxPOA which extends mqc.Servant
      //For the reply message => this object manipulate the buffer at bufferPtr location
      byte[] objectKey = msg.getObjectKey(); //Get the correct object
      int key = getObjKeyKey(objectKey);
      Servant servant = (Servant) activeObjectMap.get(new Integer(key));

      //Get the correct reply message implementation for the current protocol
      Reply replyHandler = getReplyHandler( msg );

      try
      {
         // Invoke operation
         IntHolder holder = new IntHolder( 0 );
         try
         {
            byte[] tmpBuf = servant._invoke(replyHandler, msg.getOperation(), msg.getBody().getBuffer(), holder );
            replyHandler.setReplyStatus(Reply.NO_EXCEPTION);

            // Set buffer into body of reply message
            replyHandler.getBody().setBuffer( tmpBuf );
            replyHandler.getBody().setPtr( holder.value );
         }
         catch (Exception e)
         {
            replyHandler.setReplyStatus(Reply.USER_EXCEPTION);
            // Any exception is currently mapped to an UNKNOWN exception
            // as specified by the OMG specification.
            replyHandler.setExceptionID("UNKNOWN");
            replyHandler.setExceptionMinorCode(1);
         }


         // Correspond request ID and endianess with request message
         replyHandler.setRequestID( msg.getRequestID() );
         replyHandler.setBigEndian( msg.isBigEndian() );
      }
      catch(mqc.exception.UNKNOWN e)
      {
         // Handle Exception
         e.printStackTrace();
      }

      if(msg.isOneWay())
      {
         return null;
      }
      else
      {
         return replyHandler;
      }
   }

   /**
    * Adds a servant implementation to the list of 
    * implementations in the POA. Returns the object
    * reference.
    *
    *  @param Servant The implementation object
    *
    * @return Object The object reference
    */
   public Object servant_to_reference( Servant servant )
   {
      byte[] objKey = createObjKey();

      int key = getObjKeyKey(objKey);

      //Add implementation to list of implementations
      activeObjectMap.put(new Integer(key), servant);

      //Create profile
      Profile profile = new Profile();
      profile.typeID = servant.getID();

      //Find all transports listening
      Hashtable transports = orb.getTransports();
      Enumeration e = transports.elements();
      S_Transport transport = null;
      TaggedProfile [] taggedProfile = null;
      while(e.hasMoreElements())
      {
         transport = (S_Transport) e.nextElement();
         taggedProfile = transport.getTaggedProfile();
         for (int i=0; i < taggedProfile.length; i++)
         {
            taggedProfile[i].objectKey = new byte[objKey.length];
            System.arraycopy(objKey, 0, taggedProfile[i].objectKey, 0, objKey.length);
            taggedProfile[i] = (TaggedProfile) taggedProfile[i].clone();
            profile.add(taggedProfile[i]);
         }
      }

      IOR ior = new IOR(profile);
      SDelegate delegate = new SDelegate(orb, ior);
      SObjImpl objectImpl = new SObjImpl();
      objectImpl.set_delegate(delegate);
      return objectImpl;
   }

   /**
    * Creates the object key used to uniquely identify
    * an object.
    *
    * @return byte[] The object key
    */
   public byte[] createObjKey()
   {
      // NOTE:
      // The length of the Object key must include length of poa name
      // and the length of the key. so for "rootPOA" this will be 7+4=11.
      
      //Object key consist of: Length of POA name, the poa name + key
      int poaLen = name.length();
      byte[] objKey = new byte[4 + poaLen + 4];
      objKey[0] = (byte) (poaLen >> 24);
      objKey[1] = (byte) (poaLen >> 16);
      objKey[2] = (byte) (poaLen >> 8);
      objKey[3] = (byte) poaLen;
      byte[] poaName = name.getBytes();
      System.arraycopy(poaName, 0, objKey, 4, poaLen);

      //System.out.println("POA::createObjKey -- name: " + name + "  key: " + key);

      objKey[4+poaLen] = (byte) (key >> 24);
      objKey[5+poaLen] = (byte) (key >> 16);
      objKey[6+poaLen] = (byte) (key >> 8);
      objKey[7+poaLen] = (byte) key;

      key++;
      return objKey;
   }

   /**
    * Get the unique key from inside the object key
    *
    * @return int The key
    */
   public int getObjKeyKey(byte[] objKey)
   {
      int poaLen = (int) (((objKey[0] & 0xff) << 24) |
                          ((objKey[1] & 0xff) << 16) |
                          ((objKey[2] & 0xff) << 8)  |
                          ((objKey[3] & 0xff)));

      int index = poaLen + 4;
      int key = (int) (((objKey[index++] & 0xff) << 24) |
                       ((objKey[index++] & 0xff) << 16) |
                       ((objKey[index++] & 0xff) << 8)  |
                       ((objKey[index++] & 0xff)));
      return key;
   }

   /**
    * Finds the correct reply handler
    * @param msg An Request message
    * @return ReplyHandler The reply handler
    */
   public Reply getReplyHandler(Request msg)
   {
      Reply reHand = null;
      __GIOP_PROTOCOL__(
      if (msg instanceof GIOPRequest)
      {
         reHand = (Reply) replyHandlers.get("GIOP");
         //Dynamically lower the version for this reply message if needed
         //((GIOPReply)reHand).setGenVersion(((GIOPMessage)msg).getMinorVersion());
      }
      )
      __GIOPLITE_PROTOCOL__(
      if (msg instanceof GIOPLiteRequest)
      {
         reHand = (Reply) replyHandlers.get("GIOPLite");
      }
      )
      __MQCIOP_PROTOCOL__(
      if (msg instanceof MQCIOPRequest)
      {
         reHand = (Reply) replyHandlers.get("MQCIOP");
      }
      )
      //We don't want to create message yet!
      reHand.setRequestID(msg.getRequestID());
      return reHand;
   }
}
