/* Id: HMAC.java,v 1.6 2000/02/09 20:32:58 gelderen Exp
 *
 * Copyright (C) 1995-2000 The Cryptix Foundation Limited.
 * All rights reserved.
 *
 * Use, modification, copying and distribution of this software is subject
 * the terms and conditions of the Cryptix General Licence. You should have
 * received a copy of the Cryptix General Licence along with this library;
 * if not, you can download a copy from http://www.cryptix.org/ .
 */
package mqc.security.mac;


import mqc.security.InvalidKeyException;
import mqc.security.Key;
import mqc.security.MessageDigest;
import mqc.security.NoSuchAlgorithmException;
import mqc.security.Mac;


/**
 * Abstract class implementing the methods common to an HMAC. Subclasses
 * implement specifics (just the name of MessageDigest to use + size).
 *
 * @author Jeroen C. van Gelderen (gelderen@cryptix.org)
 * @version Revision: 1.6 
 */
public class HMAC extends Mac
{
   /** Inner and outer padding constants */
   private static final byte
      IPAD = (byte)0x36,
      OPAD = (byte)0x5C;

   /** Underlying MessageDigest */
   private final MessageDigest md;

   /** Block size and length of underlying MessageDigest */
   private final int
      mdBlockSize,
      mdLen;

   /** Inner and outer IVs */
   private final byte[]
      iv_i,
      iv_o;


   /**
    * Construct an HMAC on top of a MessageDigest.
    */
   public HMAC(MessageDigest md)
   {
      this.md          = md;
      this.mdBlockSize = md.getBlockSize();
      this.mdLen       = md.getDigestLength();
      this.iv_i        = new byte[mdBlockSize];
      this.iv_o        = new byte[mdBlockSize];
   }


   /**
    * Private constructor used for cloning.
    */
   private HMAC(MessageDigest md, byte[] iv_i, byte[] iv_o)
   {
      this.md          = md;
      this.mdBlockSize = md.getBlockSize();
      this.mdLen       = md.getDigestLength();
      //BUG??--if we're cloning we need to make copies of the data
      //this.iv_i        = new byte[mdBlockSize];
      //this.iv_o        = new byte[mdBlockSize];
      //and the HMAC clone method (see below) appears to assume
      //that we keep the arrays, so...
      this.iv_i        = iv_i;
      this.iv_o        = iv_o;
   }


   /**
    * Return the length of this Mac which equals the length of the underlying
    * MessageDigest.
    */
   protected final int engineGetMacLength()
   {
      return mdLen;
   }


   protected final void engineInit(Key key)
      throws InvalidKeyException
   {
      if( !key.getFormat().equals("RAW") )
         throw new InvalidKeyException("HMAC accepts keys in 'RAW' format only.");

      // extract the keybytes
      byte[] keyBytes = key.getEncoded();

      // hash down if key > blocksize
      if( keyBytes.length > this.mdBlockSize )
      {
         md.reset();
         keyBytes = md.digest( keyBytes );
      }

      // zero both iv's
      for( int i=0; i<iv_i.length; i++ )
         iv_i[i] = iv_o[i] = 0x00;

      // copy key into iv_i and xor with IPAD
      System.arraycopy( keyBytes, 0, iv_i, 0, keyBytes.length );
      for( int i=0; i<iv_i.length; i++ )
         iv_i[i] ^= IPAD;

      // copy key into iv_o and xor with OPAD
      System.arraycopy( keyBytes, 0, iv_o, 0, keyBytes.length );
      for( int i=0; i<iv_i.length; i++ )
         iv_o[i] ^= OPAD;

      // reset the engine to initial state
      engineReset();
   }


   protected final void engineUpdate(byte input)
   {
      md.update(input);
   }


   protected final void engineUpdate(byte[] input, int offset, int len)
   {
      md.update(input, offset, len);
   }


   protected final byte[] engineDoFinal()
   {
      // FIXME: don't create temporary objects here

      // obtain the inner hash and reset underlying MessageDigest
      byte[] tmp = md.digest();
      md.reset();

      // finish mac computation by prepending the inner hash with
      // the inner IV and hashing the result
      md.update(iv_o);
      md.update(tmp);
      byte[] output = md.digest();

      // reset engine and return mac
      engineReset();
      return output;
   }


   protected final void engineReset()
   {
      // reset digest and hash in the inner IV
      md.reset();
      md.update(iv_i);
   }


   /**
    * Clone this HMAC object. We support cloning if the underlying
    * MessageDigest does support cloning.
    */
   public final Object clone()
   throws CloneNotSupportedException
   {
      MessageDigest md = (MessageDigest)this.md.clone();
      //--clone not support on CLDC, do it manually
      //return new HMAC(
      //    md,
      //    (byte[])this.iv_i.clone(),
      //    (byte[])this.iv_o.clone() );
      byte[] civ_i = new byte[iv_i.length];
      System.arraycopy(iv_i, 0, civ_i, 0, iv_i.length);
      byte[] civ_o = new byte[iv_o.length];
      System.arraycopy(iv_o, 0, civ_o, 0, iv_o.length);
      return new HMAC( md, civ_i, civ_o);
   }
}
