/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

/* $Id: CRC32.java,v 1.3 2003/09/08 18:53:04 rjohnsto Exp $
 *
 * This code is based upon code the public domain CRC32 code written
 * by Michael Lecuyer (mjl@theorem.com).
 */

package mqc.security.md;

import mqc.security.MessageDigest;
import mqc.security.DigestException;

/**
 * Implements a CRC32-based message digest algorithm in Java.
 *
 * @author  David McKinnon
 */
public final class CRC32 extends MessageDigest implements Cloneable
{

   // Static variables and constants
   //...........................................................................

   /** The size in bytes of the input block to the tranformation algorithm. */
   private static final int BLOCK_LENGTH = 4; //    = 32 / 8;

   private static int CRCTable[] = null;   // CRC Lookup table


   // Instance variables
   //...........................................................................

   /** temporary internal crc value */
   private int crc;    // currently calculated crc


   // Constructors, clone()
   //...........................................................................

   public CRC32 ()
   {
      super();
      if (CRCTable == null)
      {
         buildCRCTable();
      }
      engineReset();
   }

   /** This constructor is here to implement cloneability of this class. */
   private CRC32(CRC32 md)
   {
      this();
      crc = md.crc;
   }

   /** Returns a copy of this MD object. */
   public Object clone()
   {
      return new CRC32(this);
   }

   /**
    * Just build a plain old fashioned table based on good, old fashioned
    * values like the CRC32_POLYNOMIAL.  The table is of a fixed size.
    */
   private void buildCRCTable()
   {
      final int CRC32_POLYNOMIAL = 0xEDB88320;

      int i, j;
      int crc;

      CRCTable = new int[256];

      for (i = 0; i <= 255; i++)
      {
         crc = i;
         for (j = 8; j > 0; j--)
            if ((crc & 1) == 1)
               crc = (crc >>> 1) ^ CRC32_POLYNOMIAL;
            else
               crc >>>= 1;
         CRCTable[i] = crc;
      }
   }


   // JCE methods
   //...........................................................................

   /** returns the Digest Length (BLOCK_LENGTH) */
   protected int engineGetDigestLength()
   {
      return BLOCK_LENGTH;
   }

   /** returns the BLOCK_LENGTH */
   protected int engineGetBlockSize()
   {
      return BLOCK_LENGTH;
   }

   /**
    * Resets this object disregarding any temporary data present at the
    * time of the invocation of this call.
    */
   protected void engineReset()
   {
      crc = 0xFFFFFFFF;
   }

   /**
    * Continues a CRC32 message digest using the input byte.
    */
   public void engineUpdate (byte input)
   {
      int temp1 = crc >>> 8;
      int temp2 = CRCTable[ (crc ^ input) & 0xFF];
      crc = temp1 ^ temp2;
   }


   /**
    * CRC32 block update operation.
    * <p>
    * Continues an CRC32 message digest operation.
    *
    * @param    input    input block
    * @param    offset   start of meaningful bytes in input
    * @param    len      count of bytes in input block to consider
    */
   protected void engineUpdate (byte[] input, int offset, int len)
   {
      // make sure we don't exceed input's allocated size/length
      if (offset < 0 || len < 0 || (long)offset + len > input.length)
         throw new ArrayIndexOutOfBoundsException();

      int temp1, temp2;
      int i = offset;

      while (len-- != 0)
      {
         temp1 = crc >>> 8;
         temp2 = CRCTable[(crc ^ input[i++]) & 0xFF];
         crc = temp1 ^ temp2;
      }
   }

   /**
    * Completes the hash computation by performing final operations such
    * as padding. At the return of this engineDigest, the MD engine is
    * reset.
    *
    * @return    the array of bytes for the resulting hash value.
    */
   protected byte[] engineDigest ()
   {
      // copy the long into a 4-byte array
      byte[] result = new byte[BLOCK_LENGTH];

      // write the crc in big-endian format
      result[0] = (byte) (crc >> 24);
      result[1] = (byte) (crc >> 16);
      result[2] = (byte) (crc >>  8);
      result[3] = (byte)  crc;

      // reset the engine
      engineReset();
      return result;
   }

   /**
    * Calls engineDigest to do the actual work
    */
   protected int engineDigest(byte[] buf, int offset, int len)
   throws DigestException
   {
      if (len < BLOCK_LENGTH)
         throw new DigestException();

      byte [] eDigest = engineDigest();
      System.arraycopy( eDigest, 0, buf, offset, eDigest.length);

      return eDigest.length;
   }


   // own methods
   //...........................................................................

   // not needed, the CRC32 code is embedded in the above method bodies
}
